// Output generated from file: /home/claude/code/local/sayagata-loop/hyperbolic-sayagata-2.frag
// Created: Tue Jun 30 19:34:51 2015
// Created: Mon Dec 9 18:58:29 2013
#include "Progressive2D.frag"
#include "Complex.frag"

const float pi = 3.141592653;

// empty border around image
const float padding = 1.05;

// which pattern to tile with
// 0: triangles
// 1: Sayagata (when q = 4)
const int tiling = 1;

// recursion depth
const int depth = 12;

// number of repetitions around the loop
const int segments = 3;

// number of zig-zag steps in each loop segment
const int nDelta = 3;
// zig-zag path angles
const int delta[nDelta] = int[](0,0,1);//, 1, 2);

// hyperbolic symmetry group
// number of q-gons about each vertex (>= 5)
const int p = 5;
// basic tile shape (must be 4 for Sayagata texture)
const int q = 4;

// hyperbolic symmetry transformations
// rotations by pi k/p at each q-gon vertex, excluding the identity
const int nTransforms = q * (p - 1);

// hyperbolic fundamental triangle lengths
// distance from pi/p vertex to pi/2 vertex (opposite)
float opp = acosh((cos(pi/float(q)) + cos(pi / float(p)))/ sin(pi/float(p)));
// distance from pi/q vertex to pi/2 vertex (adjacent)
float adj = acosh(cos(pi/float(p)) / sin(pi/float(q)));
// distance from pi/p vertex to pi/q vertex (hypotenuse)
float hyp = acosh(cos(pi/float(p)) * cos(pi/float(q)) / (sin(pi/float(p)) * sin(pi/float(q))));

// hyperbolic fundamental triangle vertices
// pi/q vertex (center of q-gon)
vec2 face = vec2(0.0, 0.0);
// pi/2 vertex (center of q-gon edge)
vec2 edge = vec2(adj, 0.0);
// pi/p vertex (vertex of q-gon)
vec2 vertex = vec2(hyp * cos(pi / float(q)), hyp * sin(pi / float(q)));

// hyperbolic embedding in Poincaré disc model
// Euclidean distance
float edist(vec2 z, vec2 w) { return length(z - w); }
// hyperbolic distance
float hdist(vec2 z, vec2 w) {
  return acosh(1.0 + 2.0 * dot(z - w, z - w) / ((1.0 - dot(z, z)) * (1.0 - dot(w, w))));
}
// hyperbolic distance to Euclidean distance
float h2e(float z) { return sqrt((cosh(z) - 1.0) / (cosh(z) + 1.0)); }
vec2 h2e(vec2 z) { return normalize(z) * h2e(length(z)); }
// Euclidean distance to hyperbolic distance
float e2h(float z) { return hdist(vec2(z, 0.0), vec2(0.0, 0.0)); }
vec2 e2h(vec2 z) { return normalize(z) * e2h(length(z)); }

// homogenous coordinates for hyperbolic space
vec4 moebius(vec2 z) { return vec4(z, 1.0, 0.0); }
vec2 unmoebius(vec4 z) { return cDiv(z.xy, z.zw); }
float mlength(vec4 z) { return length(z.xy) / length(z.zw); }

// identity transformation matrix
mat4 identity() {
  return mat4
    ( 1.0, 0.0, 0.0, 0.0
    , 0.0, 1.0, 0.0, 0.0
    , 0.0, 0.0, 1.0, 0.0
    , 0.0, 0.0, 0.0, 1.0
    );
}

// rotation transformation matrix
mat4 rotation(float a) {
  float c = cos(a);
  float s = sin(a);
  return mat4
    ( c, -s, 0.0, 0.0
    , s,  c, 0.0, 0.0
    , 0.0, 0.0, 1.0, 0.0
    , 0.0, 0.0, 0.0, 1.0
    );
}

// translation transformation matrix
mat4 translation(vec2 z) {
  float s = atan(z.y, z.x);
  float l = length(z);
  float e = exp(l);
  float f = e + 1.0;
  float g = e - 1.0;
  mat4 m = mat4
    ( f, 0.0, g, 0.0
    , 0.0, f, 0.0, g
    , g, 0.0, f, 0.0
    , 0.0, g, 0.0, f
    );
  if (l > 0.0) {
    return rotation(-s) * m * rotation(s);
  } else {
    return m;
  }
}

// rotation about a point
mat4 rotationAbout(vec2 z, float a) {
  return translation(z) * rotation(a) * translation(-z);
}

// complex atan
vec2 cAtan(vec2 x) {
  vec2 one = vec2(1.0, 0.0);
  vec2 i = vec2(0.0, 1.0);
  vec2 ix = cMul(i, x);
  return cMul(0.5 * i, cLog(one - ix) - cLog(one + ix));
}

// procedural triangle tile
float triangle(vec4 m0) {
  vec2 w = unmoebius(m0);
  float a = atan(w.y, w.x) / (2.0 * pi);
  a = mod(float(q) * a, 1.0);
  return float(a > 0.5);
}

// procedural hyperbolic Sayagata tile
float sayagata(vec4 m0) {
  vec4 m = m0;
  float v = 0.0;
    vec2 w = abs(unmoebius(m));
    m = moebius(w);
    if (w.x < h2e(1.5 * adj / 9.0)) { v = 0.0; } else {
      w.x -= h2e(2.0 * adj / 9.0); w.x = abs(w.x);
      if (w.x < h2e(1.0 * adj / 9.0)) { v = 1.0; } else {
      w.x -= h2e(2.0 * adj / 9.0); w.x = abs(w.x);
      if (w.x < h2e(1.0 * adj / 9.0)) { v = 0.0; } else {
        m = rotationAbout(vertex, 2.0 * pi / float(p)) * m;
        w = abs(unmoebius(m));
         if (w.x < h2e(1.5 * adj / 9.0)) { v = 0.0; } else {
        w.x -= h2e(2.0 * adj / 9.0); w.x = abs(w.x);
        if (w.x < h2e(1.0 * adj / 9.0)) { v = 1.0; } else {
        w.x -= h2e(2.0 * adj / 9.0); w.x = abs(w.x);
        if (w.x < h2e(1.0 * adj / 9.0) && w.y < h2e(12.0 * adj / 9.0) ) { v = 0.0; } else {
        if (w.x > h2e(2.0 * adj / 9.0) && w.y > h2e(11.0 * adj / 9.0)) { v = 0.0; } else {
          v = 1.0;
     }}}}}}}
  return v;
}

// compute output colour given fragment coordinates
vec3 color(vec2 z) {

  // each segment moves deltaU squares in one direction,
  // and deltaV squares in the other direction
  // s is the vector from the orgin to its segment-equivalent q-gon
  vec4 u = moebius(vec2(0.0));
  for (int i = 0; i < nDelta; ++i) {
    float a = 2.0 * pi * float(delta[i]) / float(q);
    if ((q & 1) == 1 && (i & 1) == 1) {
      a += pi / float(q);
    }
    vec2 v = 2.0 * adj * vec2(cos(a), sin(a));
    u = translation(v) * u;
  }
  vec2 s = unmoebius(u);

//  im(arctan(x + iy)) =  -(1/4) ln ((1 - x^2 - y^2)^2 + (2x)^2) + (1/2) ln ((1 + y)^2 + x^2) 
//  im(arctan(iy)) =  -(1/4) ln ((1 - y^2)^2) + (1/2) ln ((1 + y)^2)
  // distance between centers of segment-equivalent q-gons
  float y = length(s);
  // I can't remember how I derived this formula for t and tFactor...
  float t = log(1.0 + y) - 0.5 * log(1.0 - y*y);
  float tFactor = pi / t / float(segments);

  // boundary of Poincaré disc backtransformed to image coordinates
  // this makes the outer edge of the annulus invariant over parameters
  float scaleFactor = exp(atan(1.0) * tFactor);
  z *= scaleFactor * padding;

  // compute transformation matrices
  mat4 transformation[nTransforms];
  int k = 0;
  // for each vertex
  for (int i = 0; i < q; ++i) {
    vec2 vert = vec2(hyp * cos(pi * float(2 * i + 1) / float(q)), hyp * sin(pi * float(2 * i + 1) / float(q)));
    // for each rotation
    for (int j = 1; j < p; ++j) {
      transformation[k++] = rotationAbout(vert, 2.0 * pi * float(j) / float(p));
    }
  }

  // map annulus to canonical segment
  float r = length(z);
  float a = mod(float(segments) * atan(z.y, z.x), 2.0 * pi) / float(segments);
//  a -= pi / float(segments);
  z = r * vec2(cos(a), sin(a));

  // map annulus segment to tangent strip
  z = cLog(z);
  z.y += 2.0 * pi;
  z /= tFactor;

  if (abs(z.x) >= pi / 4.0) {
    // outside tangent strip
    return vec3(1.0);
  }

  // map tangent strip to Poincaré disc
  z = cDiv(cSin(z), cCos(z));

  if (length(z) >= 1.0) {
    // outside Poincaré disc
    return vec3(1.0);
  }

  // rotate so (deltaU, deltaV) lines up properly at segment boundaries
//  vec2 s1 = translate(s) * translate(vec2(adj, 0.0)) * vec2(0.0);
  float theta = atan(s.y, s.x) + pi / 2.0;
  vec4 m = rotation(-theta) * moebius(z);

  // repeatedly apply transformations towards the fundamental domain
  for (int i = 0; i < depth; ++i) {
    if (mlength(m) <= h2e(adj)) {
       // m is within the incircle of the fundamental domain
      break;
   }
    vec4 mm = m;
    bool done = true;
    for (int j = 0; j < nTransforms; ++j) {
      vec4 mmm = transformation[j] * m;
      if (mlength(mmm) < mlength(mm)) {
        // keep the closest to the origin
        mm = mmm;
        done = false;
      }
    }
    m = normalize((mm));
    if (done) {
      // no transformation brought it closer to the origin
      break;
    }
  }

  // colour fundamental domain using procedural texture
  switch (tiling) {
  case 1:
    if (q == 4) { return vec3(1.0 - sayagata(m)); }
    break;
  }
   return vec3(triangle(m));

}


#preset default
Center = 0,0
Zoom = 1
Gamma = 2.2
ToneMapping = 1
Exposure = 1
Brightness = 1
Contrast = 1
Saturation = 1
AARange = 2
AAExp = 1
GaussianAA = true
#endpreset

